import unittest

from pydantic import ValidationError

from scipeds import constants
from scipeds.data.enums import AwardLevel, FieldTaxonomy, Grouping, RaceEthn
from scipeds.data.queries import QueryFilters, TaxonomyRollup


class QueryFilterTests(unittest.TestCase):
    def test_years(self):
        # Bad year ranges fail validation
        with self.assertRaises(ValidationError):
            QueryFilters(start_year=2000, end_year=1999)

        # Race/ethnicity change year span raises warning
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            QueryFilters(start_year=2010, end_year=2011)

            # Default query filter produces default values
            qf = QueryFilters()
        assert qf.start_year == constants.START_YEAR
        assert qf.end_year == constants.END_YEAR

        # Adding values outside the bounds won't validate
        with self.assertRaises(ValidationError):
            QueryFilters(start_year=constants.START_YEAR - 1)
        with self.assertRaises(ValidationError):
            QueryFilters(end_year=constants.END_YEAR + 1)

    def test_race_ethns(self):
        # Adding singular values yields lists
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            qf = QueryFilters(race_ethns=RaceEthn.american_indian)
        assert qf.race_ethns == [RaceEthn.american_indian]

        # Correct strings are okay
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            qf = QueryFilters(race_ethns=[RaceEthn.american_indian])
        assert qf.race_ethns == [RaceEthn.american_indian]

        # Unidentified values fail
        with self.assertRaises(ValidationError):
            QueryFilters(race_ethns=["male"])

    def test_degrees(self):
        # Adding singular values yields lists
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            qf = QueryFilters(award_levels=AwardLevel.bachelors)
        assert qf.award_levels == [AwardLevel.bachelors]

        # Correct strings are okay
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            qf = QueryFilters(race_ethns=[RaceEthn.american_indian.value])
        assert qf.race_ethns == [RaceEthn.american_indian]

        # Unidentified values fail
        with self.assertRaises(ValidationError):
            QueryFilters(race_ethns=["male"])

    def test_majornum(self):
        # Value of 3 fails
        with self.assertRaises(ValidationError):
            QueryFilters(majornums=[3])


class FieldTaxonomyTests(unittest.TestCase):
    def test_fields(self):
        # Test valid column name
        sd = TaxonomyRollup(taxonomy_name="cipcode", taxonomy_values="12.3456")
        assert sd.taxonomy_name == FieldTaxonomy.original_cip
        assert sd.taxonomy_values == ["12.3456"]

        # Test invalid column name
        with self.assertRaises(ValidationError):
            TaxonomyRollup(taxonomy_name="not_a_valid_taxonomy", taxonomy_values=[1])


class GroupingTests(unittest.TestCase):
    def test_suffix(self):
        assert Grouping.intersectional.students_suffix == "students"
        assert Grouping.gender.students_suffix == ""
        assert Grouping.race_ethnicity.students_suffix == ""

    def test_label(self):
        assert Grouping.intersectional.label_suffix == "intersectional"
        assert Grouping.gender.label_suffix == "within_gender"
        assert Grouping.race_ethnicity.label_suffix == "within_race_ethnicity"

    def test_grouping_columns(self):
        assert Grouping.intersectional.grouping_columns == ["race_ethnicity", "gender"]
        assert Grouping.gender.grouping_columns == ["gender"]
        assert Grouping.race_ethnicity.grouping_columns == ["race_ethnicity"]



import unittest
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from scipeds.constants import CIP_TABLE, COMPLETIONS_TABLE, INSTITUTIONS_TABLE
from scipeds.data.completions import CompletionsQueryEngine
from scipeds.data.enums import (
    AwardLevel,
    FieldTaxonomy,
    Grouping,
    NCSESDetailedFieldGroup,
    NCSESSciGroup,
    RaceEthn,
)
from scipeds.data.queries import QueryFilters, TaxonomyRollup


class QueryEngineTests(unittest.TestCase):
    def setUp(self):
        # We want to replicate the process for creating the duckdb as closely as we can
        # We do this by creating a fake dataset and using the pipeline to write test assets.
        # To re-generate the test db, run `make test-assets`
        test_db = Path(__file__).parents[0] / "assets" / "test.duckdb"

        self.engine = CompletionsQueryEngine(test_db)
        test_completions = self.engine.get_df_from_query(f"SELECT * FROM {COMPLETIONS_TABLE}")
        test_metadata = self.engine.get_df_from_query(
            f"SELECT DISTINCT unitid, institution_name FROM {INSTITUTIONS_TABLE}"
        )
        unis = test_metadata.drop_duplicates().set_index("unitid")
        self.names = unis.to_dict()["institution_name"]
        self.unitids = sorted(test_completions.unitid.unique())
        self.years = sorted(test_completions.year.unique())
        self.genders = sorted(test_completions.gender.unique())
        self.dfgs = sorted(test_completions[FieldTaxonomy.ncses_detailed_field_group].unique())
        self.cips = sorted(test_completions.cipcode.unique())
        self.dhs_stems = [True, False]
        self.race_ethnicities = sorted(test_completions.race_ethnicity.unique())

    def _check_result(self, result: pd.DataFrame, expected: pd.DataFrame):
        """Convenience wrapper for checking accurate query return values

        We ignore categorical values and dtypes"""
        n_cols = expected.shape[1]
        assert_frame_equal(
            result.iloc[:, :n_cols],
            expected,
            check_categorical=False,
            check_index_type=False,
        )

    def test_list_tables(self):
        """Make sure returned list of tables is correct"""
        expected_tables = [
            COMPLETIONS_TABLE,
            CIP_TABLE,
            INSTITUTIONS_TABLE,
        ]
        returned_tables = self.engine.list_tables()
        self.assertCountEqual(expected_tables, returned_tables)

    def test_cip_query(self):
        cip_table = self.engine.get_cip_table()
        assert not cip_table.empty

        # cip code is index, ncses + dhs fields are 5, title col makes 6
        assert len(cip_table.columns) == 6

        returned_cip_titles = cip_table["cip_title"].values.tolist()
        expected_cip_titles = [
            "English Literature (British And Commonwealth)",
            "Mathematics, General",
            "Behavioral Neuroscience",
        ]
        self.assertCountEqual(returned_cip_titles, expected_cip_titles)

    def test_institution_characteristics_query(self):
        inst_table = self.engine.get_institutions_table()
        assert not inst_table.empty

        returned_unitids = inst_table.index
        expected_unitids = [1, 2]
        self.assertCountEqual(returned_unitids, expected_unitids)

        inst_table = self.engine.get_institutions_table(cols=["institution_name"])
        self.assertCountEqual(inst_table.columns, ["institution_name"])
        self.assertEqual(inst_table.index.name, "unitid")

    def test_rollup_check(self):
        good_rollup = TaxonomyRollup(
            taxonomy_name=FieldTaxonomy.ncses_sci_group, taxonomy_values=NCSESSciGroup.sci
        )
        with warnings.catch_warnings():
            warnings.simplefilter("error", UserWarning)
            self.engine._check_rollup_values(good_rollup)

        bad_rollup = TaxonomyRollup(
            taxonomy_name=FieldTaxonomy.ncses_detailed_field_group,
            taxonomy_values=NCSESSciGroup.sci,
        )
        with self.assertWarns(UserWarning):
            self.engine._check_rollup_values(bad_rollup)

    def test_filters(self):
        # Filtering out bachelors should give us an empty dataset
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters(award_levels=[AwardLevel.masters])
        fields_agg = TaxonomyRollup(
            taxonomy_name=FieldTaxonomy.ncses_sci_group, taxonomy_values=NCSESSciGroup.sci
        )

        result = self.engine.rollup_by_grouping(
            grouping=Grouping.gender, rollup=fields_agg, query_filters=filters
        )
        assert result.empty

        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.gender,
            taxonomy=FieldTaxonomy.ncses_field_group,
            query_filters=filters,
        )
        assert result.empty

        # Wrong year should also give missing values
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters(start_year=1996, end_year=2021)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.gender, rollup=fields_agg, query_filters=filters
        )
        assert result.empty

        # Trying to recover race/ethnicity data prior to 1995 raises a warning
        with self.assertWarnsRegex(UserWarning, "Race/ethnicity"):
            filters = QueryFilters(
                start_year=1991, end_year=1995, race_ethns=[RaceEthn.two_or_more]
            )

        # Only looking for "unknown" race/ethnicity is okay
        with warnings.catch_warnings():
            warnings.simplefilter("error", UserWarning)
            filters = QueryFilters(start_year=1991, end_year=1995, race_ethns=[RaceEthn.unknown])

        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.gender,
            taxonomy=FieldTaxonomy.ncses_field_group,
            query_filters=filters,
        )
        assert result.empty

        # Filtering to missing race/ethnicity category should yield nothing
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters(race_ethns=[RaceEthn.unknown])
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.gender, rollup=fields_agg, query_filters=filters
        )
        assert result.empty

        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.gender,
            taxonomy=FieldTaxonomy.ncses_field_group,
            query_filters=filters,
        )
        assert result.empty

        # Missing majornums should yield nothing
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters(majornums=2)
        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.gender,
            taxonomy=FieldTaxonomy.ncses_field_group,
            query_filters=filters,
        )
        assert result.empty

    def test_group_query(self):
        # Test the basic group query
        one_stem = TaxonomyRollup(
            taxonomy_name=FieldTaxonomy.ncses_detailed_field_group,
            taxonomy_values=[NCSESDetailedFieldGroup.math_stats],
        )
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters()

        index_cols = [
            "gender",
        ]
        value_cols = [
            "rollup_degrees_within_gender",
            "rollup_degrees_total",
            "uni_degrees_within_gender",
            "uni_degrees_total",
        ]

        # One major in STEM = 3 men/women per uni in each group
        # Two unis => double
        # Two years => double again
        one_stem_gender_subtotals = np.array([3, 6, 9, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[[gender] + one_stem_gender_subtotals.tolist() for gender in self.genders],
        ).set_index(index_cols)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.gender, rollup=one_stem, query_filters=filters
        )
        self._check_result(result, expected)

        # Defining both math + psych as stem should change results
        two_stem = TaxonomyRollup(
            taxonomy_name=FieldTaxonomy.ncses_detailed_field_group,
            taxonomy_values=[NCSESDetailedFieldGroup.math_stats, NCSESDetailedFieldGroup.psych],
        )
        # Two majors in STEM = 6 men/women per group per uni
        # Two unis => double
        # Two years => double again
        two_stem_gender_subtotals = np.array([6, 12, 9, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[[gender] + two_stem_gender_subtotals.tolist() for gender in self.genders],
        ).set_index(index_cols)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.gender, rollup=two_stem, query_filters=filters
        )
        self._check_result(result, expected)

        # Test grouping by race
        index_cols = [
            "race_ethnicity",
        ]
        value_cols = [
            "rollup_degrees_within_race_ethnicity",
            "rollup_degrees_total",
            "uni_degrees_within_race_ethnicity",
            "uni_degrees_total",
        ]
        # One major in STEM = 2 people per group per uni
        # Two unis => double
        # Two years => double again
        one_stem_re_subtotals = np.array([2, 6, 6, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[[race] + one_stem_re_subtotals.tolist() for race in self.race_ethnicities],
        ).set_index(index_cols)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.race_ethnicity, rollup=one_stem, query_filters=filters
        )
        self._check_result(result, expected)

        # Two majors in STEM = 2 * 2 = 4 people per group per uni
        # Two unis => double
        # Two years => double again
        two_stem_re_subtotals = np.array([4, 12, 6, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[[race] + two_stem_re_subtotals.tolist() for race in self.race_ethnicities],
        ).set_index(index_cols)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.race_ethnicity, rollup=two_stem, query_filters=filters
        )
        self._check_result(result, expected)

        # Test intersectional
        index_cols = [
            "race_ethnicity",
            "gender",
        ]
        value_cols = [
            "rollup_degrees_intersectional",
            "rollup_degrees_total",
            "uni_degrees_intersectional",
            "uni_degrees_total",
        ]

        # Splitting by gender+race with one major means we expect one person per sub-group
        # and 3 people within that sub-group overall at each uni
        # Two unis => double
        # Two years => double again
        subtotals = np.array([1, 6, 3, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [race, gender] + subtotals.tolist()
                for race in self.race_ethnicities
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.intersectional, rollup=one_stem, query_filters=filters
        )
        self._check_result(result, expected)

        # Test by year
        # One major in STEM = 3 men/women per uni in each group
        # Two unis => double
        # one per year
        index_cols = [
            "gender",
            "year",
        ]
        value_cols = [
            "rollup_degrees_within_gender",
            "rollup_degrees_total",
            "uni_degrees_within_gender",
            "uni_degrees_total",
        ]
        one_stem_gender_subtotals = np.array([3, 6, 9, 18]) * 2 * 1
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [gender, year] + one_stem_gender_subtotals.tolist()
                for gender in self.genders
                for year in self.years
            ],
        ).set_index(index_cols)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.gender, rollup=one_stem, query_filters=filters, by_year=True
        )
        self._check_result(result, expected)

        # test unitid filter
        # One major in STEM = 3 men/women per uni in each group
        # One uni only
        # one each year
        one_stem_gender_subtotals = np.array([3, 6, 9, 18]) * 1 * 1
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [gender, year] + one_stem_gender_subtotals.tolist()
                for gender in self.genders
                for year in self.years
            ],
        ).set_index(index_cols)
        result = self.engine.rollup_by_grouping(
            grouping=Grouping.gender,
            rollup=one_stem,
            query_filters=filters,
            by_year=True,
            filter_unitids=[1],
        )
        self._check_result(result, expected)

    def test_group_fields_query(self):
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters()
        index_cols = [
            "ncses_detailed_field_group",
            "gender",
        ]
        value_cols = [
            "field_degrees_within_gender",
            "field_degrees_total",
            "uni_degrees_within_gender",
            "uni_degrees_total",
        ]
        # 3 person per major per gender
        # 6 people per major total
        # 9 people per uni degree per gender
        # 18 people per uni degree total
        # 2 unis => double
        # 2 years => double
        subtotals = np.array([3, 6, 9, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [dfg, gender] + subtotals.tolist() for dfg in self.dfgs for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.gender,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
        )
        self._check_result(result, expected)

        # Test race
        index_cols = [
            "ncses_detailed_field_group",
            "race_ethnicity",
        ]
        value_cols = [
            "field_degrees_within_race_ethnicity",
            "field_degrees_total",
            "uni_degrees_within_race_ethnicity",
            "uni_degrees_total",
        ]
        # 2 peple per major per race
        # 6 people per race subtotal
        # 6 people per uni per race
        # 18 people per uni total
        # 2 unis => double
        # 2 years => double
        subtotals = np.array([2, 6, 6, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [dfg, race] + subtotals.tolist()
                for dfg in self.dfgs
                for race in self.race_ethnicities
            ],
        ).set_index(index_cols)
        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.race_ethnicity,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
        )
        self._check_result(result, expected)

        # Test by year
        index_cols = [
            "ncses_detailed_field_group",
            "race_ethnicity",
            "year",
        ]
        value_cols = [
            "field_degrees_within_race_ethnicity",
            "field_degrees_total",
            "uni_degrees_within_race_ethnicity",
            "uni_degrees_total",
        ]
        # 2 peple per major per race
        # 6 people per race subtotal
        # 6 people per uni per race
        # 18 people per uni total
        # 2 unis => double
        # each year
        subtotals = np.array([2, 6, 6, 18]) * 2 * 1
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [dfg, race, year] + subtotals.tolist()
                for dfg in self.dfgs
                for race in self.race_ethnicities
                for year in self.years
            ],
        ).set_index(index_cols)
        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.race_ethnicity,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
            by_year=True,
        )
        self._check_result(result, expected)

        # Intersectional
        index_cols = [
            "ncses_detailed_field_group",
            "race_ethnicity",
            "gender",
        ]
        value_cols = [
            "field_degrees_intersectional",
            "field_degrees_total",
            "uni_degrees_intersectional",
            "uni_degrees_total",
        ]
        # 1 person per major per race+gender
        # 6 people per major per total
        # 3 people per uni per race+gender
        # 18 people per uni total
        # 2 unis => double
        # 2 years => double
        subtotals = np.array([1, 6, 3, 18]) * 2 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [field, race, gender] + subtotals.tolist()
                for field in self.dfgs
                for race in self.race_ethnicities
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.intersectional,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
        )
        self._check_result(result, expected)

        # Test unitid filter
        # 1 person per major per race+gender
        # 6 people per major per total
        # 3 people per uni per race+gender
        # 18 people per uni total
        # 1 uni
        # 2 years => double
        subtotals = np.array([1, 6, 3, 18]) * 1 * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [field, race, gender] + subtotals.tolist()
                for field in self.dfgs
                for race in self.race_ethnicities
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.field_totals_by_grouping(
            grouping=Grouping.intersectional,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
            filter_unitids=[2],
        )
        self._check_result(result, expected)

    def test_uni_query(self):
        # Test gender
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters()
        fields_agg = TaxonomyRollup(
            taxonomy_name=FieldTaxonomy.ncses_sci_group, taxonomy_values=NCSESSciGroup.sci
        )
        index_cols = [
            "institution_name",
            "unitid",
            "gender",
        ]
        value_cols = [
            "rollup_degrees_within_gender",
            "rollup_degrees_total",
            "uni_degrees_within_gender",
            "uni_degrees_total",
        ]
        # 3 people per major within gender per uni
        # 2 timesteps
        subtotals = np.array([3, 6, 9, 18]) * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, gender] + subtotals.tolist()
                for unitid in self.unitids
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.uni_rollup_by_grouping(
            grouping=Grouping.gender, rollup=fields_agg, query_filters=filters
        )
        self._check_result(result, expected)

        # By year
        index_cols = [
            "institution_name",
            "unitid",
            "gender",
            "year",
        ]
        value_cols = [
            "rollup_degrees_within_gender",
            "rollup_degrees_total",
            "uni_degrees_within_gender",
            "uni_degrees_total",
        ]
        # 3 people per major within gender per uni
        subtotals = np.array([3, 6, 9, 18])
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, gender, year] + subtotals.tolist()
                for unitid in self.unitids
                for gender in self.genders
                for year in self.years
            ],
        ).set_index(index_cols)
        result = self.engine.uni_rollup_by_grouping(
            grouping=Grouping.gender, rollup=fields_agg, query_filters=filters, by_year=True
        )
        self._check_result(result, expected)

        # Test race/ethn
        index_cols = [
            "institution_name",
            "unitid",
            "race_ethnicity",
        ]
        value_cols = [
            "rollup_degrees_within_race_ethnicity",
            "rollup_degrees_total",
            "uni_degrees_within_race_ethnicity",
            "uni_degrees_total",
        ]
        # 2 people per major per race/ethn per uni
        # 1 STEM major of 3 total majors
        # 2 timesteps
        subtotals = np.array([2, 6, 6, 18]) * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, race_ethn] + subtotals.tolist()
                for unitid in self.unitids
                for race_ethn in self.race_ethnicities
            ],
        ).set_index(index_cols)
        result = self.engine.uni_rollup_by_grouping(
            grouping=Grouping.race_ethnicity, rollup=fields_agg, query_filters=filters
        )
        self._check_result(result, expected)

        # Test intersectional
        index_cols = [
            "institution_name",
            "unitid",
            "race_ethnicity",
            "gender",
        ]
        value_cols = [
            "rollup_degrees_intersectional",
            "rollup_degrees_total",
            "uni_degrees_intersectional",
            "uni_degrees_total",
        ]
        # 2 people per major per race/ethn per uni
        # 1 STEM major of 3 total majors
        # 2 timesteps
        subtotals = np.array([1, 6, 3, 18]) * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, race_ethn, gender] + subtotals.tolist()
                for unitid in self.unitids
                for race_ethn in self.race_ethnicities
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.uni_rollup_by_grouping(
            grouping=Grouping.intersectional, rollup=fields_agg, query_filters=filters
        )
        self._check_result(result, expected)

        # Test unitid filter
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, race_ethn, gender] + subtotals.tolist()
                for unitid in [1]
                for race_ethn in self.race_ethnicities
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.uni_rollup_by_grouping(
            grouping=Grouping.intersectional,
            rollup=fields_agg,
            query_filters=filters,
            filter_unitids=[1],
        )
        self._check_result(result, expected)

    def test_uni_fields_query(self):
        # Test gender
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters()
        index_cols = [
            "institution_name",
            "unitid",
            "ncses_detailed_field_group",
            "gender",
        ]
        value_cols = [
            "field_degrees_within_gender",
            "field_degrees_total",
            "uni_degrees_within_gender",
            "uni_degrees_total",
        ]
        # 3 people per major within gender per uni
        # 2 timesteps
        subtotals = np.array([3, 6, 9, 18]) * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, dfg, gender] + subtotals.tolist()
                for unitid in self.unitids
                for dfg in self.dfgs
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.uni_field_totals_by_grouping(
            grouping=Grouping.gender,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
        )
        self._check_result(result, expected)

        # Test by year
        index_cols = [
            "institution_name",
            "unitid",
            "ncses_detailed_field_group",
            "gender",
            "year",
        ]
        # 3 people per major within gender per uni
        subtotals = np.array([3, 6, 9, 18])
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, dfg, gender, year] + subtotals.tolist()
                for unitid in self.unitids
                for dfg in self.dfgs
                for gender in self.genders
                for year in self.years
            ],
        ).set_index(index_cols)
        result = self.engine.uni_field_totals_by_grouping(
            grouping=Grouping.gender,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
            by_year=True,
        )
        self._check_result(result, expected)

        # Test race/ethn
        index_cols = [
            "institution_name",
            "unitid",
            "ncses_detailed_field_group",
            "race_ethnicity",
        ]
        value_cols = [
            "field_degrees_within_race_ethnicity",
            "field_degrees_total",
            "uni_degrees_within_race_ethnicity",
            "uni_degrees_total",
        ]
        # 2 people per major per race/ethn per uni
        # 1 STEM major of 3 total majors
        # 2 timesteps
        subtotals = np.array([2, 6, 6, 18]) * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, dfg, race_ethn] + subtotals.tolist()
                for unitid in self.unitids
                for dfg in self.dfgs
                for race_ethn in self.race_ethnicities
            ],
        ).set_index(index_cols)
        result = self.engine.uni_field_totals_by_grouping(
            grouping=Grouping.race_ethnicity,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
        )
        self._check_result(result, expected)

        # Test intersectional
        index_cols = [
            "institution_name",
            "unitid",
            "ncses_detailed_field_group",
            "race_ethnicity",
            "gender",
        ]
        value_cols = [
            "field_degrees_intersectional",
            "field_degrees_total",
            "uni_degrees_intersectional",
            "uni_degrees_total",
        ]
        # 2 people per major per race/ethn per uni
        # 1 STEM major of 3 total majors
        # 2 timesteps
        subtotals = np.array([1, 6, 3, 18]) * 2
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, dfg, race_ethn, gender] + subtotals.tolist()
                for unitid in self.unitids
                for dfg in self.dfgs
                for race_ethn in self.race_ethnicities
                for gender in self.genders
            ],
        ).set_index(index_cols)
        result = self.engine.uni_field_totals_by_grouping(
            grouping=Grouping.intersectional,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            query_filters=filters,
        )
        self._check_result(result, expected)

        # Test intersectional by year with filtered fields
        index_cols = [
            "institution_name",
            "unitid",
            "ncses_detailed_field_group",
            "race_ethnicity",
            "gender",
            "year",
        ]
        value_cols = [
            "field_degrees_intersectional",
            "field_degrees_total",
            "uni_degrees_intersectional",
            "uni_degrees_total",
        ]
        # 2 people per major per race/ethn per uni
        # 1 STEM major of 3 total majors
        # 2 timesteps
        subtotals = np.array([1, 6, 3, 18])
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, dfg, race_ethn, gender, year] + subtotals.tolist()
                for unitid in self.unitids
                for dfg in self.dfgs[:2]
                for race_ethn in self.race_ethnicities
                for gender in self.genders
                for year in self.years
            ],
        ).set_index(index_cols)
        result = self.engine.uni_field_totals_by_grouping(
            grouping=Grouping.intersectional,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            taxonomy_values=self.dfgs[:2],
            query_filters=filters,
            by_year=True,
        )
        self._check_result(result, expected)

        # Test unitid filter
        subtotals = np.array([1, 6, 3, 18])
        expected = pd.DataFrame(
            columns=index_cols + value_cols,
            data=[
                [self.names[unitid], unitid, dfg, race_ethn, gender, year] + subtotals.tolist()
                for unitid in [2]
                for dfg in self.dfgs
                for race_ethn in self.race_ethnicities
                for gender in self.genders
                for year in self.years
            ],
        ).set_index(index_cols)
        result = self.engine.uni_field_totals_by_grouping(
            grouping=Grouping.intersectional,
            taxonomy=FieldTaxonomy.ncses_detailed_field_group,
            taxonomy_values=self.dfgs,
            query_filters=filters,
            by_year=True,
            filter_unitids=[2],
        )
        self._check_result(result, expected)

    def test_rel_rate(self):
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters()
        for grouping in Grouping:
            for taxonomy in FieldTaxonomy:
                for yearly in (True, False):
                    if taxonomy == FieldTaxonomy.original_cip:
                        continue
                    result = self.engine.field_totals_by_grouping(
                        grouping=grouping,
                        taxonomy=taxonomy,
                        query_filters=filters,
                        by_year=yearly,
                        rel_rate=True,
                    )
                    assert "rel_rate" in result.columns

    def test_effect_size(self):
        with self.assertWarnsRegex(UserWarning, "IPEDS"):
            filters = QueryFilters()
        taxonomies = [
            FieldTaxonomy.ncses_detailed_field_group,
            FieldTaxonomy.dhs_stem,
            FieldTaxonomy.cip,
        ]
        values = [self.dfgs, self.dhs_stems, self.cips]

        for grouping in Grouping:
            for taxonomy, vals in zip(taxonomies, values):
                for yearly in (True, False):
                    rollup = TaxonomyRollup(taxonomy_name=taxonomy.value, taxonomy_values=vals)
                    result = self.engine.uni_rollup_by_grouping(
                        grouping=grouping,
                        rollup=rollup,
                        query_filters=filters,
                        by_year=yearly,
                        rel_rate=True,
                        effect_size=True,
                    )
                    assert "excess_degrees_from_median_expected" in result.columns

        for grouping in Grouping:
            for taxonomy in FieldTaxonomy:
                if taxonomy == FieldTaxonomy.original_cip:
                    continue
                for yearly in (True, False):
                    result = self.engine.uni_field_totals_by_grouping(
                        grouping=grouping,
                        taxonomy=taxonomy,
                        query_filters=filters,
                        by_year=yearly,
                        rel_rate=True,
                        effect_size=True,
                    )
                    assert "excess_degrees_from_median_expected" in result.columns



import warnings
from typing import Any, List

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from typing_extensions import Annotated

from scipeds import constants
from scipeds.data.enums import AwardLevel, FieldTaxonomy, RaceEthn
from scipeds.utils import validate_and_listify


class TaxonomyRollup(BaseModel):
    """Defining values within a taxonomy to include in an aggregation"""

    model_config = ConfigDict(extra="forbid")

    taxonomy_name: FieldTaxonomy
    """The taxonomy within which to aggregate over values"""

    taxonomy_values: List[Any]
    """The values in the taxonomy to include in aggregation"""

    @field_validator("taxonomy_values", mode="before")
    @classmethod
    def values_valid_list(cls, v: Any):
        if not isinstance(v, list):
            v = [v]
        return v


class QueryFilters(BaseModel):
    """A pydantic model with options for how to filter the baseline data prior to any aggregation.

    The model will handle data validation for the input values by ensuring:

    - The years are in range and the range is nonzero
    - The race/ethnicity and degree values provided are valid
    - Things that ought to be lists are lists
    """

    model_config = ConfigDict(extra="forbid")

    start_year: Annotated[int, Field(ge=constants.START_YEAR, le=constants.END_YEAR)] = (
        constants.START_YEAR
    )
    """The start year of the time window to aggregate over, inclusive"""

    end_year: Annotated[int, Field(ge=constants.START_YEAR, le=constants.END_YEAR)] = (
        constants.END_YEAR
    )
    """The end year of the time window to aggregate over, inclusive"""

    race_ethns: List[RaceEthn] = list(RaceEthn)
    """ Which race/ethnicity groups to include (default: all)"""

    award_levels: List[AwardLevel] = list(AwardLevel)
    """Which degrees / award levels to include (default: all)"""

    majornums: List[Annotated[int, Field(ge=1, le=2)]] = [1, 2]
    """Which major numbers to include (default: both first and second majors)"""

    @field_validator("race_ethns", mode="before")
    @classmethod
    def race_ethn_valid(cls, v: Any):
        # Validate provided race/ethnicity values, turn into a list
        return validate_and_listify(v, RaceEthn)

    @field_validator("award_levels", mode="before")
    @classmethod
    def award_valid(cls, v: Any):
        # Validate provided degree values, turn into a list
        return validate_and_listify(v, AwardLevel)

    @field_validator("majornums", mode="before")
    @classmethod
    def majornum_valid(cls, v: Any):
        # Validate provided majornum values, turn into a list
        if isinstance(v, int):
            v = [v]
        return v

    @model_validator(mode="after")
    def check_year_range(self):
        """Validate year values

        Warns:
            UserWarning: If time range overlaps significant changes in IPEDS schema or values

        Raises:
            ValueError: If provided start_year is after end_year
        """
        if self.start_year > self.end_year:
            raise ValueError("Start year must be less than or equal to end year")
        # Warn if time range overlaps span where racial categorization changes
        if self.start_year <= 2010 and self.end_year >= 2011:
            warnings.warn(
                "IPEDS award level coding and race and ethnicity coding changed "
                "between 2010 and 2011 datasets. For more details, see "
                "https://nces.ed.gov/ipeds/report-your-data/race-ethnicity-reporting-changes"
            )
        if (self.start_year < 1995 or self.end_year < 1995) and self.race_ethns != [
            RaceEthn.unknown
        ]:
            warnings.warn(
                "Race/ethnicity data is not available before 1995. All race/ethnicities "
                "prior to 1995 are treated as 'unknown'."
            )
        return self



import numpy as np
from numpy.testing import assert_array_almost_equal

from scipeds.utils import (
    bounded_ratio_transform,
    forward_fold_transform,
    inverse_bounded_ratio_transform,
    inverse_fold_transform,
)


def test_forward_fold_transform():
    """Test the forward fold transform

    Expected behavior
        # 1 -> 0      # Parity
        # 0.5 -> -1   # 2x under-represented
        # 3 -> 2      # 3x over-represented
        # 0 -> -inf   # completely underrepresented
        # inf -> inf  # completely overrepresented
    """
    values = np.array([1, 0.5, 3, 0, np.inf])
    expected = np.array([0, -1, 2, -np.inf, np.inf])
    transformed = forward_fold_transform(values)
    assert_array_almost_equal(expected, transformed)


def test_inverse_fold_transform():
    """Test the inverse fold transform

    Expected behavior
        # 0 -> 1      # Parity
        # -1 -> 0.5   # 2x under-represented
        # 2 -> 3      # 3x over-represented
        # -inf -> 0   # completely underrepresented
        # inf -> inf  # completely overrepresented
    """
    values = np.array([0, -1, 2, -np.inf, np.inf])
    expected = np.array([1, 0.5, 3, 0, np.inf])
    transformed = inverse_fold_transform(values)
    assert_array_almost_equal(expected, transformed)


def test_bounded_transform():
    """Test the bounded ratio transform

    Expected behavior
        1 -> 0        # Parity
        0.5 -> -0.5   # 2x under-represented
        3 -> 0.66666  # 3x over-represented
        0 -> -1       # completely underrepresented
        inf -> 1      # completely overrepresented
    """
    values = np.array([1, 0.5, 3, 0, np.inf])
    expected = np.array([0, -0.5, 2 / 3, -1, 1])
    transformed = bounded_ratio_transform(values)
    assert_array_almost_equal(expected, transformed)


def test_inverse_bounded_transform():
    """Test the bounded ratio transform

    Expected behavior
        0 -> 1        # Parity
        -0.5 -> 0.5   # 2x under-represented
        0.66666 -> 3  # 3x over-represented
        -1 -> 0       # completely under-represented
        1 -> inf      # completely over-represented
    """
    values = np.array([0, -0.5, 2 / 3, -1, 1])
    expected = np.array([1, 0.5, 3, 0, np.inf])
    transformed = inverse_bounded_ratio_transform(values)
    assert_array_almost_equal(expected, transformed)



import unittest
from itertools import chain

from scipeds.data.enums import (
    NCSES_HIERARCHY,
    NCSESDetailedFieldGroup,
    NCSESFieldGroup,
    NCSESSciGroup,
)


class TestFields(unittest.TestCase):
    def test_sci_group(self):
        self.assertCountEqual(list(NCSES_HIERARCHY.keys()), list(NCSESSciGroup))

    def test_field_group(self):
        fields = list(
            chain.from_iterable(list(NCSES_HIERARCHY[sg].keys()) for sg in NCSESSciGroup)
        )
        self.assertCountEqual(fields, list(NCSESFieldGroup))

    def test_detailed_field_group(self):
        detailed_fields = list(
            chain.from_iterable(
                list(
                    chain.from_iterable(list(NCSES_HIERARCHY[sg].values())) for sg in NCSESSciGroup
                )
            )
        )
        self.assertCountEqual(detailed_fields, list(NCSESDetailedFieldGroup))



import warnings

import pandas as pd

from scipeds.constants import COMPLETIONS_TABLE, INSTITUTIONS_TABLE
from scipeds.data.engine import IPEDSQueryEngine
from scipeds.data.enums import FieldTaxonomy, Grouping
from scipeds.data.queries import QueryFilters, TaxonomyRollup
from scipeds.utils import (
    Rate,
    calculate_effect_size,
    calculate_rel_rate,
)


class CompletionsQueryEngine(IPEDSQueryEngine):
    GROUP_FIELDS_QUERY = """
WITH filtered AS (
    SELECT * FROM {completions_table}
    WHERE 
        year BETWEEN $start_year AND $end_year
        AND race_ethnicity IN (SELECT UNNEST($race_ethns))
        AND awlevel IN (SELECT UNNEST($award_levels))
        AND majornum IN (SELECT UNNEST($majornums))
        {unitid_filter}
), 
field_group_totals AS (
    {field_group_total_select}
), 
field_totals AS (
    {field_total_select}
), 
group_totals AS (
    {group_total_select}
), 
totals AS (
    {total_select}
)
SELECT
    {institution_name}
    {field_group_cols},
    {field_group_degrees},
    {field_total_degrees},
    {group_total_degrees},
    {total_degrees},
FROM field_group_totals
{joins}
ORDER BY {field_group_cols};"""

    def _check_rollup_values(self, rollup: TaxonomyRollup):
        """Check whether the rollup values provided exist in the specified taxonomy, and warn
        the user if they do not.

        Args:
            rollup (TaxonomyRollup): Taxonomy rollup
        """
        unique_values = self.get_df_from_query(
            f"SELECT DISTINCT {rollup.taxonomy_name} FROM {COMPLETIONS_TABLE}"
        )[rollup.taxonomy_name]
        missing = [value for value in rollup.taxonomy_values if value not in unique_values.values]
        if (n_missing := len(missing)) > 0:
            warnings.warn(
                f"{n_missing} provided taxonomy values were missing from taxonomy "
                f"{rollup.taxonomy_name}. First {n_missing if n_missing < 3 else 3} "
                f"missing values: {list(missing)[:3]}.\n"
                "The query will still proceed but beware that you might not get "
                "the correct or expected results."
            )

    def _format_query(
        self,
        grouping: Grouping,
        agg_type: str,
        field_group_cols: list[str],
        field_total_cols: list[str],
        group_total_cols: list[str],
        total_cols: list[str],
        taxonomy: str | None = None,
        unitids: list[int] | None = None,
    ) -> str:
        """Format the query according to the provided arguments

        Args:
            grouping (Grouping): How to group the data
            field_group_cols (list[str]): Lowest level of aggregation
            field_total_cols (list[str]): Field-only aggregation
            group_total_cols (list[str]): Grouping-only aggregation
            total_cols (list[str]): Highest level of aggregation
            taxonomy (str, Optional): Taxonomy to filter on. Default: None
            unitids (list[int], Optional): List of unitids to filter on. Default: None

        Returns:
            str: Formatted query
        """
        taxonomy_filter = (
            f"WHERE {taxonomy} IN (SELECT UNNEST($taxonomy_values))" if taxonomy else ""
        )
        unitid_filter = (
            f"AND unitid IN ({', '.join([str(unitid) for unitid in unitids])})" if unitids else ""
        )

        def format_subquery(cols: list[str], name: str, filter: str = "") -> str:
            selection = ", ".join(cols) + "," if cols else ""
            groupby = f"GROUP BY {', '.join(cols)}" if cols else ""
            return (
                f"SELECT {selection} "
                f"  COALESCE(SUM(n_awards), 0)::INT64 AS {name} "
                f"FROM filtered {filter} "
                f"{groupby}"
            )

        field_group_total_select = format_subquery(
            field_group_cols, f"{agg_type}_degrees_{grouping.label_suffix}", taxonomy_filter
        )
        field_total_select = format_subquery(
            field_total_cols, f"{agg_type}_degrees_total", taxonomy_filter
        )
        group_total_select = format_subquery(
            group_total_cols, f"uni_degrees_{grouping.label_suffix}"
        )
        total_select = format_subquery(total_cols, "uni_degrees_total")

        def format_join(cols: list[str], var_name: str, table_name: str) -> tuple[str, str]:
            if cols:
                join_str = f"LEFT JOIN {table_name} USING ({', '.join(cols)})"
            else:
                var_name = f"{table_name}.{var_name}"
                join_str = f",{table_name}"
            return var_name, join_str

        fgt_var, _ = format_join(
            field_group_cols, f"{agg_type}_degrees_{grouping.label_suffix}", "field_group_totals"
        )
        ft_var, ft_join = format_join(
            field_total_cols, f"{agg_type}_degrees_total", "field_totals"
        )
        gt_var, gt_join = format_join(
            group_total_cols, f"uni_degrees_{grouping.label_suffix}", "group_totals"
        )
        t_var, t_join = format_join(total_cols, "uni_degrees_total", "totals")
        all_joins = list(sorted([ft_join, gt_join, t_join], reverse=True))
        if "unitid" in total_cols:
            all_joins.append(f"LEFT JOIN {INSTITUTIONS_TABLE} USING (unitid)")
        joins = "\n".join(all_joins)

        institution_name = (
            f"{INSTITUTIONS_TABLE}.institution_name," if "unitid" in total_cols else ""
        )

        query = self.GROUP_FIELDS_QUERY.format(
            completions_table=COMPLETIONS_TABLE,
            institution_name=institution_name,
            unitid_filter=unitid_filter,
            field_group_cols=", ".join(field_group_cols),
            field_group_total_select=field_group_total_select,
            field_total_select=field_total_select,
            group_total_select=group_total_select,
            total_select=total_select,
            field_group_degrees=fgt_var,
            field_total_degrees=ft_var,
            group_total_degrees=gt_var,
            total_degrees=t_var,
            joins=joins,
        )

        return query

    def rollup_by_grouping(
        self,
        grouping: Grouping,
        rollup: TaxonomyRollup,
        query_filters: QueryFilters,
        by_year: bool = False,
        rel_rate: bool = False,
        show_query: bool = False,
        filter_unitids: list[int] | None = None,
    ) -> pd.DataFrame:
        """Aggregate completions (subject to filters) for fields within the given roll-up,
        aggregating by selected grouping and subject to the applied filters

        Args:
            grouping (Grouping): How to group the data
            rollup (TaxonomyRollup): Fields in taxonomy to include in aggregation
            query_filters (QueryFilters): Filters to apply prior to aggregation
            by_year (bool): Whether to group by year (True) or aggregate over all years (False).
                Default: False
            rel_rate (bool): Whether to calculate relative representation. If true,
                also adds associated variables. Default: False
            show_query (bool): Whether to print the query and parameters before executing.
                Default: False
        Returns:
            pd.DataFrame: Completions within fields in the roll-up,
                aggregated by chosen grouping and subject to filters
        """
        # Validate inputs
        grouping = Grouping(grouping)

        # Warn the user if rollup values are missing from the specified taxonomy
        self._check_rollup_values(rollup)

        year = ["year"] if by_year else []
        field_group_cols = [*grouping.grouping_columns, *year]
        field_total_cols = [*year]
        group_total_cols = [*grouping.grouping_columns, *year]
        total_cols = [*year]

        # Format and execute the query
        query = self._format_query(
            grouping=grouping,
            agg_type="rollup",
            field_group_cols=field_group_cols,
            field_total_cols=field_total_cols,
            group_total_cols=group_total_cols,
            total_cols=total_cols,
            taxonomy=rollup.taxonomy_name,
            unitids=filter_unitids,
        )
        query_params = query_filters.model_dump()
        query_params.update(rollup.model_dump(include={"taxonomy_values"}))
        df = self.get_df_from_query(query, query_params=query_params, show_query=show_query)

        # Calculate relative rate
        if rel_rate:
            rollup_pct = Rate(
                "rollup_pct", f"rollup_degrees_{grouping.label_suffix}", "rollup_degrees_total"
            )
            uni_pct = Rate("uni_pct", f"uni_degrees_{grouping.label_suffix}", "uni_degrees_total")
            df = calculate_rel_rate(df, rollup_pct, uni_pct)

        return df.set_index(field_group_cols)

    def field_totals_by_grouping(
        self,
        grouping: Grouping,
        taxonomy: FieldTaxonomy,
        query_filters: QueryFilters,
        taxonomy_values: list[str] | None = None,
        by_year: bool = False,
        rel_rate: bool = False,
        show_query: bool = False,
        filter_unitids: list[int] | None = None,
    ) -> pd.DataFrame:
        """Compute aggregate counts for all fields in a given taxonomy

        Args:
            grouping (Grouping): How to group the data
            taxonomy (FieldTaxonomy): Taxonomy to aggregate over
            query_filters (QueryFilters): Pre-aggregation filters to apply to raw data
            by_year (bool): Whether to group by year (True) or aggregate over all years (False).
                Default: False
            rel_rate (bool): Whether to calculate relative representation. Default: False
            show_query (bool): Whether to print the query and parameters before executing.
                Default: False

        Returns:
            pd.DataFrame: Relative rates by grouping for each field in taxonomy
        """
        # Validate inputs
        grouping = Grouping(grouping)
        taxonomy = FieldTaxonomy(taxonomy)

        year = ["year"] if by_year else []
        field_group_cols = [taxonomy.value, *grouping.grouping_columns, *year]
        field_total_cols = [taxonomy.value, *year]
        group_total_cols = [*grouping.grouping_columns, *year]
        total_cols = [*year]

        query = self._format_query(
            grouping=grouping,
            agg_type="field",
            field_group_cols=field_group_cols,
            field_total_cols=field_total_cols,
            group_total_cols=group_total_cols,
            total_cols=total_cols,
            taxonomy=taxonomy if taxonomy_values else None,
            unitids=filter_unitids,
        )
        query_params = query_filters.model_dump()
        df = self.get_df_from_query(query, query_params=query_params, show_query=show_query)

        if rel_rate:
            # Calculate relative rate
            field_pct = Rate(
                "field_pct", f"field_degrees_{grouping.label_suffix}", "field_degrees_total"
            )
            uni_pct = Rate("uni_pct", f"uni_degrees_{grouping.label_suffix}", "uni_degrees_total")
            df = calculate_rel_rate(df, field_pct, uni_pct)

        return df.set_index(field_group_cols)

    def uni_rollup_by_grouping(
        self,
        grouping: Grouping,
        rollup: TaxonomyRollup,
        query_filters: QueryFilters,
        by_year: bool = False,
        rel_rate: bool = False,
        effect_size: bool = False,
        show_query: bool = False,
        filter_unitids: list[int] | None = None,
    ) -> pd.DataFrame:
        """Get intersectional degree counts and rates within intersectional subgroups"

        Args:
            grouping (Grouping): How to group the data
            rollup (TaxonomyRollup): Taxonomy to aggregate over
            query_filters (QueryFilters): Pre-aggregation filters
            by_year (bool): Whether to group by year (True) or aggregate over all years (False).
                Default: False
            rel_rate (bool): Whether to calculate relative representation. If true,
                also adds associated variables. Default: False
            effect_size (bool): Whether to compute effect size. Default: False
            show_query (bool): Whether to print the query and parameters before executing.
                Default: False

        Returns:
            pd.DataFrame: Completions in fields contained within roll-up, aggregated by
                university UNITID and chosen grouping, subject to filters
        """
        # Validate inputs
        grouping = Grouping(grouping)

        # Warn the user if rollup values are missing from the specified taxonomy
        self._check_rollup_values(rollup)

        year = ["year"] if by_year else []
        field_group_cols = ["unitid", *grouping.grouping_columns, *year]
        field_total_cols = ["unitid", *year]
        group_total_cols = ["unitid", *grouping.grouping_columns, *year]
        total_cols = ["unitid", *year]

        query = self._format_query(
            grouping=grouping,
            agg_type="rollup",
            field_group_cols=field_group_cols,
            field_total_cols=field_total_cols,
            group_total_cols=group_total_cols,
            total_cols=total_cols,
            taxonomy=rollup.taxonomy_name,
            unitids=filter_unitids,
        )
        query_params = query_filters.model_dump()
        query_params.update(rollup.model_dump(include={"taxonomy_values"}))
        df = self.get_df_from_query(query, query_params=query_params, show_query=show_query)

        rollup_pct = Rate(
            "rollup_pct", f"rollup_degrees_{grouping.label_suffix}", "rollup_degrees_total"
        )
        uni_pct = Rate("uni_pct", f"uni_degrees_{grouping.label_suffix}", "uni_degrees_total")

        if rel_rate or effect_size:
            # Calculate z-score for rollup field pct relative to baseline uni pct
            df = calculate_rel_rate(df, rollup_pct, uni_pct)

        if effect_size:
            df = calculate_effect_size(df, rollup_pct, uni_pct, group_cols=field_group_cols[1:])

        return df.set_index(["institution_name"] + field_group_cols)

    def uni_field_totals_by_grouping(
        self,
        grouping: Grouping,
        taxonomy: FieldTaxonomy,
        query_filters: QueryFilters,
        taxonomy_values: list[str] | None = None,
        by_year: bool = False,
        rel_rate: bool = False,
        effect_size: bool = False,
        show_query: bool = False,
        filter_unitids: list[int] | None = None,
    ) -> pd.DataFrame:
        """Aggregate completions (subject to filters) for all fields
        within a given taxonomy at each university

        Args:
            grouping (Grouping): How to group the data
            taxonomy (FieldTaxonomy): Taxonomy to aggregate over
            query_filters (QueryFilters): Pre-aggregation filters to apply to raw data
            taxonomy_values (list[str]): Optional list of field values to filter on. Default: None
            by_year (bool): Whether to group by year (True) or aggregate over all years (False).
                Default: False
            rel_rate (bool): Whether to calculate relative representation. If true,
                also adds associated variables. Default: False
            effect_size (bool): Whether to compute effect size. Default: False
            show_query (bool): Whether to print the query and parameters before executing.
                Default: False

        Returns:
            pd.DataFrame: Completions in each field in the taxonomy, aggregated by
                university UNITID and chosen grouping, subject to filters
        """
        # Validate inputs
        grouping = Grouping(grouping)
        taxonomy = FieldTaxonomy(taxonomy)

        year = ["year"] if by_year else []
        field_group_cols = ["unitid", taxonomy.value, *grouping.grouping_columns, *year]
        field_total_cols = ["unitid", taxonomy.value, *year]
        group_total_cols = ["unitid", *grouping.grouping_columns, *year]
        total_cols = ["unitid", *year]

        query = self._format_query(
            grouping=grouping,
            agg_type="field",
            field_group_cols=field_group_cols,
            field_total_cols=field_total_cols,
            group_total_cols=group_total_cols,
            total_cols=total_cols,
            taxonomy=taxonomy if taxonomy_values else None,
            unitids=filter_unitids,
        )

        query_params = query_filters.model_dump()
        if taxonomy_values:
            query_params["taxonomy_values"] = taxonomy_values

        df = self.get_df_from_query(query, query_params=query_params, show_query=show_query)

        field_pct = Rate(
            "field_pct", f"field_degrees_{grouping.label_suffix}", "field_degrees_total"
        )
        uni_pct = Rate("uni_pct", f"uni_degrees_{grouping.label_suffix}", "uni_degrees_total")

        if rel_rate or effect_size:
            df = calculate_rel_rate(df, field_pct, uni_pct)

        if effect_size:
            df = calculate_effect_size(df, field_pct, uni_pct, group_cols=field_group_cols[1:])

        return df.set_index(["institution_name"] + field_group_cols)



from pathlib import Path
from typing import Any, Dict, List, Optional

import duckdb
import pandas as pd

from scipeds.constants import CIP_TABLE, DB_NAME, INSTITUTIONS_TABLE, SCIPEDS_CACHE_DIR


class IPEDSQueryEngine:
    def __init__(self, db_path: Optional[Path] = SCIPEDS_CACHE_DIR / DB_NAME):
        """A structured way to query the IPEDS table to format data for visualization

        Args:
            db_path (Optional[Path], optional): Path to pre-processed database file.
                Defaults to CACHE_DIR / DB_NAME.

        Raises:
            FileNotFoundError: Pre-processed database file not found.
        """
        if db_path and not db_path.exists():
            raise FileNotFoundError(
                f"No db file found at {db_path}! "
                "Specify an existing pre-processed db file "
                "or run scipeds.download_db() to download."
            )

        self.db_path = str(db_path)

    def get_df_from_query(
        self, query: str, query_params: Optional[Dict[str, Any]] = None, show_query: bool = False
    ) -> pd.DataFrame:
        """Return the dataframe result of the provided SQL query on the pre-processed duckdb

        Args:
            query (str): SQL query (using duckdb syntax)
            query_params (Dict[str, Any], optional): Prepared statement variables for query.
                Defaults to None.
            show_query (bool): Whether to print the query and parameters before executing.
                Defaults to False

        Returns:
            pd.DataFrame: Data returned by query
        """
        if show_query:
            print(f"Query: {query}, Params: {query_params}")
        with duckdb.connect(self.db_path, read_only=True) as con:
            if query_params is not None:
                df = con.execute(query, query_params).df()
            else:
                df = con.sql(query).df()
        return df

    def list_tables(self) -> List[str]:
        """List all tables in the duckdb

        Returns:
            List[str]: A list of all available tables
        """
        tables = self.get_df_from_query("SHOW TABLES").iloc[:, 0].values.tolist()
        # Temporary numpy typing fix - see https://github.com/numpy/numpy/issues/27944
        return tables  # type: ignore[return-value]

    def get_cip_table(self) -> pd.DataFrame:
        """Get a table of every unique 2020 CIP Code

        Returns:
            pd.DataFrame: Data frame of CIP codes and corresponding taxonomy titles
        """
        cip_codes = self.get_df_from_query(
            f"SELECT * FROM {CIP_TABLE} ORDER BY cip2020"
        ).set_index("cip2020")
        return cip_codes

    def get_institutions_table(self, cols: str | list[str] | None = None) -> pd.DataFrame:
        """Get institution characteristics table, optionally with specified columns

        Returns:
            pd.DataFrame: Data frame of institution characteristics
        """
        inst_df = self.get_df_from_query(f"SELECT * FROM {INSTITUTIONS_TABLE}").set_index("unitid")
        if isinstance(cols, str):
            cols = [cols]
        if cols is None:
            return inst_df
        for col in inst_df.columns:
            assert col in inst_df.columns, f"Invalid column name(s) provided: {cols}"
        return inst_df[cols]



import unittest

import pandas as pd

from scipeds.data.enums import Gender
from scipeds.utils import Rate, calculate_effect_size, calculate_rel_rate


class CalculationTests(unittest.TestCase):
    def setUp(self):
        """Create a fake result aggregation with three schools:
        one big school, one medium-size school, one small school

        Big school is over-represented in a field by 1.1x, small school is under-represented by 2x,
        medium school is at parity
        Use gender, for now
        TODO: Add tests for other fields / groupings"""
        unitids = [1, 2, 3]
        names = ["Big School", "Medium School", "Small School"]
        gender = Gender.women
        uni_degrees_total = [10_000, 1_000, 100]
        uni_pct_gender = [0.4, 0.5, 0.6]
        uni_degrees_within_gender = [
            int(uni_degrees * pct) for uni_degrees, pct in zip(uni_degrees_total, uni_pct_gender)
        ]
        field_pct_uni = [0.2, 0.15, 0.25]
        field_degrees_total = [
            int(uni_degrees * field_pct)
            for uni_degrees, field_pct in zip(uni_degrees_total, field_pct_uni)
        ]

        self.rel_rate = [0.8, 1, 1.2]
        field_pct_gender = [uni_pct * rr for uni_pct, rr in zip(uni_pct_gender, self.rel_rate)]

        field_degrees_within_gender = [
            int(field_degrees * pct)
            for field_degrees, pct in zip(field_degrees_total, field_pct_gender)
        ]

        self.parity_effect_size = [
            fdwg - int(fdt * upg)
            for fdwg, fdt, upg in zip(
                field_degrees_within_gender, field_degrees_total, uni_pct_gender
            )
        ]

        self.result_df = pd.DataFrame.from_dict(
            {
                "institution_name": names,
                "unitid": unitids,
                "gender": [gender.value] * 3,
                "rollup_degrees_within_gender": field_degrees_within_gender,
                "rollup_degrees_total": field_degrees_total,
                "uni_degrees_within_gender": uni_degrees_within_gender,
                "uni_degrees_total": uni_degrees_total,
            },
        )

        self.rollup_pct = Rate(
            "rollup_pct",
            numerator="rollup_degrees_within_gender",
            denominator="rollup_degrees_total",
        )
        self.uni_pct = Rate(
            "uni_pct", numerator="uni_degrees_within_gender", denominator="uni_degrees_total"
        )

    def test_relative_rate(self):
        rr_df = calculate_rel_rate(
            df=self.result_df, subgroup=self.rollup_pct, baseline=self.uni_pct
        )
        assert all(col in rr_df.columns for col in ["rel_rate", "log2_rel_rate", "fold_rel_rate"])

        actual_values = rr_df["rel_rate"].tolist()
        for actual, expected in zip(actual_values, self.rel_rate):
            self.assertAlmostEqual(actual, expected)

    def test_effect_size(self):
        rr_df = calculate_rel_rate(
            df=self.result_df, subgroup=self.rollup_pct, baseline=self.uni_pct
        )
        es_df = calculate_effect_size(
            df=rr_df,
            field_pct=self.rollup_pct,
            uni_pct=self.uni_pct,
            group_cols=["gender"],
        )

        assert all(
            col in es_df.columns
            for col in ["excess_degrees_from_parity", "excess_degrees_from_median_expected"]
        )

        # median rel rate should be 1
        assert (es_df["median_rel_rate_uni"] == 1).all()

        actual_values = es_df["excess_degrees_from_parity"].tolist()
        for actual, expected in zip(actual_values, self.parity_effect_size):
            self.assertAlmostEqual(actual, expected)

        actual_values = es_df["excess_degrees_from_median_expected"].tolist()
        for actual, expected in zip(actual_values, self.parity_effect_size):
            self.assertAlmostEqual(actual, expected)



from collections import namedtuple
from typing import Any, Iterable, List, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
from statsmodels.stats import proportion  # type: ignore

Rate = namedtuple("Rate", ["name", "numerator", "denominator"])


def forward_fold_transform(values: npt.NDArray | pd.Series) -> npt.NDArray:
    """Transform values from raw relative rate values to a linear fold scale.

    Relative rates are asymmetric around parity and on the half-open interval [0, +np.inf).
    This function transforms relative rates into a linear scale of the multiplicative value
    indicated by the rate so that the scale is symmetric around zero (parity) and on the interval
    (-np.inf, np.inf).

    Examples:
        1 -> 0  # Parity
        0.5 -> -1  # 2x under-represented
        3 -> 2  # 3x over-represented
        0 -> -inf # completely underrepresented
        inf -> inf # completely overrepresented

    Note that the numeric values on a linear fold scale differ from the labeled values by 1.
    """
    values = values.copy()
    null_idx = np.flatnonzero(values <= 0)
    values[null_idx] = np.nan
    transformed = np.array(list(map(lambda x: 1 - np.divide(1, x) if x <= 1 else x - 1, values)))
    transformed[null_idx] = -np.inf
    return transformed


def inverse_fold_transform(values: npt.NDArray | pd.Series) -> npt.NDArray:
    """Transform values from a linear fold scale to raw value"""
    return np.array(list(map(lambda x: -1 / (x - 1) if x < 0 else x + 1, values)))


def bounded_ratio_transform(values: npt.NDArray | pd.Series) -> npt.NDArray:
    """Transform values from raw relative rate values to a bounded ratio scale.

    Relative rates are asymmetric around parity and on the half-open interval [0, +np.inf).
    For certain kinds of clustering algorithms, it's useful for these to be on a symmetric,
    bounded interval.

    This function transforms the half-open interval to a fully closed interval [-1, 1]
    where 0 is parity, -1 is completely under-represented, and 1 is completely over-represented.

    Examples:
        1 -> 0  # Parity
        0.5 -> -0.5  # 2x under-represented
        3 -> 0.66666  # 3x over-represented
        0 -> -1 # completely underrepresented
        inf -> 1 # completely overrepresented
    """
    values = values.copy()
    transformed = np.array(list(map(lambda x: 1 - np.divide(1, x) if x >= 1 else x - 1, values)))
    return transformed


def inverse_bounded_ratio_transform(values: npt.NDArray | pd.Series) -> npt.NDArray:
    """Transform values from a bounded ratio scale to a raw relative rate.

    Examples:
        0 -> 1  # Parity
        -0.5 -> 0.5  # 2x under-represented
        2/3 -> 3  # 3x over-represented
        -1 -> 0 # completely underrepresented
        1 -> inf # completely overrepresented
    """
    values = values.copy()

    def _invert(x):
        if x <= 0:
            return x + 1
        if x < 1:
            return 1 / (1 - x)
        return np.inf

    transformed = np.array(list(map(_invert, values)))
    return transformed


def calculate_rel_rate(df: pd.DataFrame, subgroup: Rate, baseline: Rate) -> pd.DataFrame:
    """Calculate multiple kinds of relative rates given a subgroup and a baseline Rate"""
    df = df.copy()
    df[subgroup.name] = df[subgroup.numerator] / df[subgroup.denominator]
    df[baseline.name] = df[baseline.numerator] / df[baseline.denominator]
    df["rel_rate"] = df[subgroup.name] / df[baseline.name]
    df["log2_rel_rate"] = df["rel_rate"].apply(np.log2)
    df["fold_rel_rate"] = forward_fold_transform(df["rel_rate"])
    df["bounded_rel_rate"] = bounded_ratio_transform(df["rel_rate"])

    return df


def calculate_effect_size(
    df: pd.DataFrame, field_pct: Rate, uni_pct: Rate, group_cols: List[str]
) -> pd.DataFrame:
    """Compute the number of degrees expected if a university's relative rate in a given field
    were the same as the median relative rate in that field among all universities
    """
    df = df.copy()
    # Calculate distance from parity (and z-score)
    df["excess_degrees_from_parity"] = df[field_pct.numerator] - (
        df[uni_pct.name] * df[field_pct.denominator]
    )
    df[["excess_degrees_from_parity_zscore", "excess_degrees_from_parity_pvalue"]] = get_zscore(
        df, field_pct, uni_pct
    )
    df["median_rel_rate_uni"] = df.groupby(group_cols, observed=True).rel_rate.transform(
        lambda x: x.median()
    )
    df["median_rel_rate_expected_degrees"] = (
        df[field_pct.denominator] * df[uni_pct.name] * df["median_rel_rate_uni"]
    )
    # Avoid the scenario where the number of expected degrees is larger than the field size
    df["median_rel_rate_expected_degrees"] = df[
        ["median_rel_rate_expected_degrees", field_pct.denominator]
    ].min(axis=1)
    df["excess_degrees_from_median_expected"] = (
        df[field_pct.numerator] - df["median_rel_rate_expected_degrees"]
    )
    expected_pct = Rate(
        "median_rel_rate_expected_pct", "median_rel_rate_expected_degrees", field_pct.denominator
    )
    df[
        [
            "excess_degrees_from_median_expected_zscore",
            "excess_degrees_from_median_expected_pvalue",
        ]
    ] = get_zscore(df, field_pct, expected_pct)
    return df


def get_zscore(df: pd.DataFrame, subgroup: Rate, baseline: Rate) -> float:
    """Calculate test statistic using 2-sample test for proportions"""
    group_count, group_obs = df[subgroup.numerator], df[subgroup.denominator]
    base_count, base_obs = df[baseline.numerator], df[baseline.denominator]
    with np.errstate(divide="ignore", invalid="ignore"):
        res = proportion.test_proportions_2indep(
            group_count,
            group_obs,
            base_count,
            base_obs,
            method="agresti-caffo",
            compare="diff",
        )
    return pd.Series([res.statistic, res.pvalue])


def convert_to_series(s: Union[str, Iterable]) -> pd.Series:
    """Convert input to a pandas Series

    Args:
        s (Union[str, List[str], pd.Series]): Desired input string(s)
    """
    if isinstance(s, str):
        s = [s]
    if isinstance(s, list) or isinstance(s, np.ndarray):
        s = pd.Series(s)
    if isinstance(s, pd.Index):
        s = s.to_series()
    if isinstance(s, pd.Series):
        return s.copy()
    raise ValueError(f"Provided value {s} must be str, List[str], pd.Series, or pd.Index")


def clean_name(s: str):
    """Cleans a column name by removing punctuation and replacing whitespace with underscores"""
    s = s.replace("'", "")
    for symbol in "-:,()/":
        s = s.replace(symbol, " ")
    return "_".join(s.lower().split())


def validate_and_listify(input: Any, cls: Iterable):
    """Convenience function to validate values are members of an Enum and ensure list returned"""
    if isinstance(input, str):
        input = [input]
    elif not isinstance(input, list):
        input = list(input)
    allowed_values = set(cls)
    for value in input:
        if value not in allowed_values:
            raise ValueError(f"Value {value} not a valid member of {cls}")
    return input



import os
from pathlib import Path

from platformdirs import user_cache_path

from scipeds import __version__

# DB attributes
VERSION_STR = __version__.replace(".", "_")  # type: ignore[has-type]
DB_NAME = f"scipeds_{VERSION_STR}.duckdb"
COMPLETIONS_TABLE = "ipeds_completions_a"
INSTITUTIONS_TABLE = "ipeds_directory_info"
CIP_TABLE = "cip_info"

# Start and end year for analysis
START_YEAR = 1984
END_YEAR = 2023

# Storage bucket
SCIPEDS_BUCKET = "scipeds-data"

# Cache dir
SCIPEDS_CACHE_DIR = Path(
    os.environ.get("SCIPEDS_CACHE_DIR", (user_cache_path() / ".scipeds").resolve())
)



from pathlib import Path
from shutil import rmtree

import typer
from cloudpathlib.exceptions import CloudPathNotExistsError
from cloudpathlib.gs import GSClient
from google.cloud.storage.client import Client as StorageClient  # type: ignore

from scipeds.constants import DB_NAME, SCIPEDS_BUCKET, SCIPEDS_CACHE_DIR

app = typer.Typer()


@app.command()
def download_db(
    output_path: Path = SCIPEDS_CACHE_DIR / DB_NAME,
    overwrite: bool = False,
    verbose: bool = True,
):
    """Downloads a pre-processed duckdb file to the local machine.

    Args:
        output_path (Path, optional): Path to save db.
            Defaults to `.scipeds/ipeds.duckdb` in the user's cache path.
        overwrite (bool, optional): Whether to re-download and
            overwrite the cached db if it already exists.
            Defaults to False.
        verbose (bool, optional): Whether to verbosely log. Defaults to True.
    """
    if output_path is None:
        if verbose:
            print(
                f"No download path specified, defaulting to user cache directory "
                f"{SCIPEDS_CACHE_DIR}."
            )
        output_path = SCIPEDS_CACHE_DIR / DB_NAME
    if not (parent := output_path.parents[0]).exists():
        if verbose:
            print(f"Creating scipeds cache directory {parent.resolve()}.")
        parent.mkdir(parents=True)
    if output_path.exists() and not overwrite:
        print(
            f"Database already downloaded to {output_path}. To re-download and "
            "overwrite the existing file, re-run with `overwrite` set to `True`"
            "(from within Python) or add --overwrite to your CLI command."
        )
        return

    anonymous_client = StorageClient.create_anonymous_client()
    gs_client = GSClient(storage_client=anonymous_client)
    processed_db = gs_client.GSPath(f"gs://{SCIPEDS_BUCKET}/processed/{DB_NAME}")  # type: ignore

    if verbose:
        print(f"Downloading pre-processed IPEDS db to {output_path}")

    try:
        processed_db.download_to(output_path)
    except CloudPathNotExistsError:
        raise CloudPathNotExistsError(
            f"{processed_db} not found by scipeds. If you think it should exist, "
            "please file an issue."
        )

    if verbose:
        print("Download complete.")
        show_cache_usage(verbose)


@app.command()
def show_cache_usage(verbose: bool = True) -> float:
    """Reports how much storage is being used by scipeds"""
    sizes = [f.stat().st_size for f in SCIPEDS_CACHE_DIR.glob("**/*") if f.is_file()]
    usage_mb = sum(sizes) // 1e6
    count = len(sizes)
    if verbose:
        print(f"scipeds has {count} files and is using {usage_mb:,.0f}MB in {SCIPEDS_CACHE_DIR}")
    return usage_mb


@app.command()
def clean_cache(verbose: bool = True):
    """Removes cached db file if it exists"""
    if not SCIPEDS_CACHE_DIR.exists():
        print(f"Cache directory {SCIPEDS_CACHE_DIR} does not exist.")
        return 0
    if verbose:
        print(f"Cleaning cache directory {SCIPEDS_CACHE_DIR}")

    show_cache_usage(verbose=True)
    typer.confirm("Are you sure you want to delete the cache?", abort=True)

    rmtree(SCIPEDS_CACHE_DIR)

    if verbose:
        print(f"Cleaned cache directory {SCIPEDS_CACHE_DIR}")


if __name__ == "__main__":
    app()



from enum import Enum


class Gender(str, Enum):
    """Enumeration for gender.

    Note that [IPEDS reports gender as binary](
    https://surveys.nces.ed.gov/ipeds/public/survey-materials/faq?faqid=11).

    Attributes:
        men (str): men
        women (str): women
    """

    men = "men"
    women = "women"


class RaceEthn(str, Enum):
    """Enumeration for race/ethnicity

    Note that the categories "Two or more races" and "Native Hawaiian or Other Pacific Islander"
    were added for enrolllment data beginning in 2010-2011 and in completions data beginning in
    2011-2012. `scipeds` will warn users if their selected `QueryFilters` include years
    with mixed categories.

    See [IPEDS](https://nces.ed.gov/ipeds/report-your-data/race-ethnicity-reporting-changes)
    for more details.

    Attributes:
        american_indian (str): American Indian or Alaska Native
        asian (str): Asian
        black_or_aa (str): Black or African American
        hispanic (str): Hispanic or Latino
        hawaiian_pi (str): Native Hawaiian or Other Pacific Islander
        nonres (str): Non-resident alien
        two_or_more (str): Two or more races
        white (str): White
        unknown (str): Unknown
        total (str): Total
    """

    american_indian = "American Indian or Alaska Native"
    asian = "Asian"
    black_or_aa = "Black or African American"
    hispanic = "Hispanic or Latino"
    hawaiian_pi = "Native Hawaiian or Other Pacific Islander"
    nonres = "Non-resident alien"
    two_or_more = "Two or more races"
    white = "White"
    unknown = "Unknown"


class Grouping(str, Enum):
    gender = "gender"
    race_ethnicity = "race_ethnicity"
    intersectional = "intersectional"

    @property
    def label_suffix(self) -> str:
        return f"within_{self.value}" if self is not self.intersectional else self.value

    @property
    def students_suffix(self) -> str:
        return "students" if self is self.intersectional else ""

    @property
    def grouping_columns(self) -> list[str]:
        return (
            [self.value]
            if self is not self.intersectional
            else [self.race_ethnicity.value, self.gender.value]  # type: ignore
        )


class AwardLevel(str, Enum):
    """Enumeration for award level

    Note that in 2011, award level codes changed. The `scipeds` pipeline attempts
    to transform old codes in accordance with [IPEDS guidance](
    https://nces.ed.gov/ipeds/report-your-data/data-tip-sheet-reporting-graduate-awards).

    Attributes:
        lt_12w (str): Certificates of less than 12 weeks
        gt_12w_lt_1y (str): Certificates of at least 12 weeks but less than 1 year
        lt1 (str): Award of less than 1 academic year
        gt1_lt2 (str): Award of at least 1 but less than 2 academic years
        associates (str): Associate's degree
        gt2_lt4 (str): Award of at least 2 but less than 4 academic years
        bachelors (str): Bachelor's degree
        postbac (str): Postbaccalaureate certificate
        masters (str): Master's degree
        postmas (str): Post-master's certificates
        doctor_research (str): Doctor's degree - research/scholarship
        doctor_professional (str): Doctor's degree - professional practice
        doctor_other (str): Doctor's degree - professional practice
        unknown (str): Unknown
    """

    lt_12w = "Certificates of less than 12 weeks"
    gt_12w_lt_1y = "Certificates of at least 12 weeks but less than 1 year"
    lt1 = "Award of less than 1 academic year"
    gt1_lt2 = "Award of at least 1 but less than 2 academic years"
    associates = "Associate's degree"
    gt2_lt4 = "Award of at least 2 but less than 4 academic years"
    bachelors = "Bachelor's degree"
    postbac = "Postbaccalaureate certificate"
    masters = "Master's degree"
    postmas = "Post-master's certificates"
    doctor_research = "Doctor's degree - research/scholarship"
    doctor_professional = "Doctor's degree - professional practice"
    doctor_other = "Doctor's degree - professional practice"
    unknown = "Unknown"


class InstitutionProperty(str, Enum):
    tech_school = "Tech school"
    health_school = "Health school"
    carnegie_classification_2021_basic = "Carnegie classification"
    geographic_region = "Geographic region"
    historically_black_college_or_university = "HBCU"


class FieldTaxonomy(str, Enum):
    """Enumeration for field taxonomy.

    The values in this enumeration also correspond to columns in the pre-processed database

    Attributes:
        cip (str): The 2020 CIP code
        original_cip (str): The originally recorded CIP code
        ncses_sci_group (str): The NCSES [Science & Engineering Alternative Classification](
            https://ncsesdata.nsf.gov/sere/2018/html/sere18-dt-taba001.html)
        ncses_field_group (str): The NCSES [Broad Field Alternative Classification](
            https://ncsesdata.nsf.gov/sere/2018/html/sere18-dt-taba001.html)
        ncses_detailed_field_group (str): The NCSES
            [Detailed Field Alternative Classification](
            https://ncsesdata.nsf.gov/sere/2018/html/sere18-dt-taba001.html)
        nsf_broad_field (str): The NSF Broad Field Classification in the
            [NSF Diversity in STEM report](
            https://ncses.nsf.gov/pubs/nsf23315/report)
        dhs_stem (str): The US Department of Homeland Security
            [STEM Designated Degree Program List](
            https://www.ice.gov/doclib/sevis/pdf/stemList2024.pdf)
    """

    cip = "cip2020"
    original_cip = "cipcode"
    ncses_sci_group = "ncses_sci_group"
    ncses_field_group = "ncses_field_group"
    ncses_detailed_field_group = "ncses_detailed_field_group"
    nsf_broad_field = "nsf_broad_field"
    dhs_stem = "dhs_stem"

    def __format__(self, fmt):
        # Python 3.11 treatment of StrEnum requires overwriting formatting
        return self.value


class NCSESSciGroup(str, Enum):
    """Field values for the [NCSES Science & Engineering Alternate Classification](
            https://ncsesdata.nsf.gov/sere/2018/html/sere18-dt-taba001.html)

    Attributes:
        sci (str): Science and engineering
        non_sci (str): Non-science and engineering
        unknown (str): Not categorized in NCSES crosswalk

    """

    sci = "Science and engineering"
    non_sci = "Non-science and engineering"
    unknown = "Not categorized in NCSES crosswalk"


class NCSESFieldGroup(str, Enum):
    """Field values for the [NCSES Broad Field Alternate Classification](
            https://ncsesdata.nsf.gov/sere/2018/html/sere18-dt-taba001.html)

    Attributes:
        business_and_mgmt (str): Business and Management
        education (str): Education
        sci_eng_technologies (str): Science and Engineering Technologies
        social_sciences (str): Social Sciences
        nonsci_other (str): Other Non-sciences or Unknown Disciplines
        art_music (str): Arts and Music
        humanities (str): Humanities
        math_cs (str): Math and Computer Sciences
        nonsci_life_sci (str): Life Sciences (Non-S&E)
        sci_life_sci (str): Life Sciences
        home_ec (str): Vocational Studies and Home Economics
        engineering (str): Engineering
        phys_sci (str): Physical Sciences
        communication_librarianship (str): Communication and Librarianship
        psych (str): Psychology
        religion (str): Religion and Theology
        interdisc (str): Interdisciplinary or Other Sciences
        social_service (str): Social Service Professions
        geosci (str): Geosciences
        arch_dsgn (str): Architecture and Environmental Design
        law (str): Law
        unknown (str): Not categorized in NCSES crosswalk

    """

    business_and_mgmt = "Business and Management"
    education = "Education"
    sci_eng_technologies = "Science and Engineering Technologies"
    social_sciences = "Social Sciences"
    nonsci_other = "Other Non-sciences or Unknown Disciplines"
    art_music = "Arts and Music"
    humanities = "Humanities"
    math_cs = "Math and Computer Sciences"
    nonsci_life_sci = "Life Sciences (Non-S&E)"
    sci_life_sci = "Life Sciences"
    home_ec = "Vocational Studies and Home Economics"
    engineering = "Engineering"
    phys_sci = "Physical Sciences"
    communication_librarianship = "Communication and Librarianship"
    psych = "Psychology"
    religion = "Religion and Theology"
    interdisc = "Interdisciplinary or Other Sciences"
    social_service = "Social Service Professions"
    geosci = "Geosciences"
    arch_dsgn = "Architecture and Environmental Design"
    law = "Law"
    unknown = "Not categorized in NCSES crosswalk"


class NCSESDetailedFieldGroup(str, Enum):
    """Field values for the [NCSES Detailed Field Alternate Classification](
            https://ncsesdata.nsf.gov/sere/2018/html/sere18-dt-taba001.html)

    Attributes:
        business_and_mgmt (str): Business and Management
        nonsci_edu (str): Non-Science Education
        nonsci_other (str): Other Non-sciences or Unknown Disciplines
        art_music (str): Arts and Music
        life_sci_other (str): Other Life Sciences
        health_technologies (str): Health Technologies
        compsci (str): Computer Science
        bio (str): Biological Sciences
        home_ec (str): Vocational Studies and Home Economics
        engineering_technologies (str): Engineering Technologies
        languages (str): Foreign Languages
        communication_librarianship (str): Communication and Librarianship
        psych (str): Psychology
        polisci (str): Political Science and Public Administration
        english_lit (str): English and Literature
        math_stats (str): Mathematics and Statistics
        ag_sci (str): Agricultural Sciences
        religion (str): Religion and Theology
        social_sci_other (str): Other Social Sciences
        medical_sci (str): Medical Sciences
        history (str): History
        engineering_other (str): Other Engineering
        chem (str): Chemistry
        area_ethnic (str): Area and Ethnic Studies
        sociology (str): Sociology
        humanities_other (str): Other Humanities
        interdisc (str): Interdisciplinary or Other Sciences
        econ (str): Economics
        physics (str): Physics
        social_service (str): Social Service Professions
        electrical_eng (str): Electrical Engineering
        science_edu (str): Science Education
        earth_sci (str): Earth Sciences
        arch_dsgn (str): Architecture and Environmental Design
        law (str): Law
        civil_engineering (str): Civil Engineering
        anthro (str): Anthropology
        phys_sci_other (str): Other Physical Sciences
        mech_e (str): Mechanical Engineering
        math_edu (str): Mathematics Education
        sci_tech_edu (str): Other Science/Technical Education
        chem_e (str): Chemical Engineering
        industrial_eng (str): Industrial Engineering
        socsci_edu (str): Social Science Education
        sci_technologies (str): Science Technologies
        linguistics (str): Linguistics
        materials_engineering (str): Materials Engineering
        astronomy (str): Astronomy
        aerospace_eng (str): Aerospace Engineering
        atmospheric_sciences (str): Atmospheric Sciences
        oceanography (str): Oceanography
        sci_eng_tech_other (str): Other Science and Engineering Technologies
        hist_sci (str): History of Science
        unknown (str): Not categorized in NCSES crosswalk

    """

    business_and_mgmt = "Business and Management"
    nonsci_edu = "Non-Science Education"
    nonsci_other = "Other Non-sciences or Unknown Disciplines"
    art_music = "Arts and Music"
    life_sci_other = "Other Life Sciences"
    health_technologies = "Health Technologies"
    compsci = "Computer Science"
    bio = "Biological Sciences"
    home_ec = "Vocational Studies and Home Economics"
    engineering_technologies = "Engineering Technologies"
    languages = "Foreign Languages"
    communication_librarianship = "Communication and Librarianship"
    psych = "Psychology"
    polisci = "Political Science and Public Administration"
    english_lit = "English and Literature"
    math_stats = "Mathematics and Statistics"
    ag_sci = "Agricultural Sciences"
    religion = "Religion and Theology"
    social_sci_other = "Other Social Sciences"
    medical_sci = "Medical Sciences"
    history = "History"
    engineering_other = "Other Engineering"
    chem = "Chemistry"
    area_ethnic = "Area and Ethnic Studies"
    sociology = "Sociology"
    humanities_other = "Other Humanities"
    interdisc = "Interdisciplinary or Other Sciences"
    econ = "Economics"
    physics = "Physics"
    social_service = "Social Service Professions"
    electrical_eng = "Electrical Engineering"
    science_edu = "Science Education"
    earth_sci = "Earth Sciences"
    arch_dsgn = "Architecture and Environmental Design"
    law = "Law"
    civil_engineering = "Civil Engineering"
    anthro = "Anthropology"
    phys_sci_other = "Other Physical Sciences"
    mech_e = "Mechanical Engineering"
    math_edu = "Mathematics Education"
    sci_tech_edu = "Other Science/Technical Education"
    chem_e = "Chemical Engineering"
    industrial_eng = "Industrial Engineering"
    socsci_edu = "Social Science Education"
    sci_technologies = "Science Technologies"
    linguistics = "Linguistics"
    materials_engineering = "Materials Engineering"
    astronomy = "Astronomy"
    aerospace_eng = "Aerospace Engineering"
    atmospheric_sciences = "Atmospheric Sciences"
    oceanography = "Oceanography"
    sci_eng_tech_other = "Other Science and Engineering Technologies"
    hist_sci = "History of Science"
    unknown = "Not categorized in NCSES crosswalk"


class NSFBroadField(str, Enum):
    """Field values according to the [NSF Diversity in STEM Report](
            https://ncses.nsf.gov/pubs/nsf23315/report)

    Attributes:
        ag_and_bio_sci (str): Agricultural and biological sciences
        math_cs (str): Mathematical and computer sciences
        phys_earth_sci (str): Physical and earth sciences
        soc_behav_sci (str): Social and behavioral sciences
        eng (str): Engineering
        non_stem (str): Non-science and engineering
        unknown (str): Not categorized in NSF broad field classification
    """

    ag_and_bio_sci = "Agricultural and biological sciences"
    math_cs = "Mathematical and computer sciences"
    phys_earth_sci = "Physical and earth sciences"
    soc_behav_sci = "Social and behavioral sciences"
    eng = "Engineering"
    non_stem = "Non-science and engineering"
    unknown = "Not categorized in NSF broad field classification"


# Hierarchical relationships in the NCSES Alternative Classification
NCSES_HIERARCHY = {
    NCSESSciGroup.unknown: {NCSESFieldGroup.unknown: [NCSESDetailedFieldGroup.unknown]},
    NCSESSciGroup.non_sci: {
        NCSESFieldGroup.arch_dsgn: [NCSESDetailedFieldGroup.arch_dsgn],
        NCSESFieldGroup.art_music: [NCSESDetailedFieldGroup.art_music],
        NCSESFieldGroup.business_and_mgmt: [NCSESDetailedFieldGroup.business_and_mgmt],
        NCSESFieldGroup.communication_librarianship: [
            NCSESDetailedFieldGroup.communication_librarianship
        ],
        NCSESFieldGroup.education: [
            NCSESDetailedFieldGroup.nonsci_edu,
            NCSESDetailedFieldGroup.math_edu,
            NCSESDetailedFieldGroup.sci_tech_edu,
            NCSESDetailedFieldGroup.science_edu,
            NCSESDetailedFieldGroup.socsci_edu,
        ],
        NCSESFieldGroup.humanities: [
            NCSESDetailedFieldGroup.english_lit,
            NCSESDetailedFieldGroup.languages,
            NCSESDetailedFieldGroup.humanities_other,
            NCSESDetailedFieldGroup.history,
        ],
        NCSESFieldGroup.interdisc: [NCSESDetailedFieldGroup.interdisc],
        NCSESFieldGroup.law: [NCSESDetailedFieldGroup.law],
        NCSESFieldGroup.nonsci_life_sci: [
            NCSESDetailedFieldGroup.medical_sci,
            NCSESDetailedFieldGroup.life_sci_other,
        ],
        NCSESFieldGroup.nonsci_other: [NCSESDetailedFieldGroup.nonsci_other],
        NCSESFieldGroup.religion: [NCSESDetailedFieldGroup.religion],
        NCSESFieldGroup.sci_eng_technologies: [
            NCSESDetailedFieldGroup.engineering_technologies,
            NCSESDetailedFieldGroup.sci_technologies,
            NCSESDetailedFieldGroup.sci_eng_tech_other,
            NCSESDetailedFieldGroup.health_technologies,
        ],
        NCSESFieldGroup.home_ec: [NCSESDetailedFieldGroup.home_ec],
        NCSESFieldGroup.social_service: [NCSESDetailedFieldGroup.social_service],
    },
    NCSESSciGroup.sci: {
        NCSESFieldGroup.engineering: [
            NCSESDetailedFieldGroup.engineering_other,
            NCSESDetailedFieldGroup.aerospace_eng,
            NCSESDetailedFieldGroup.chem_e,
            NCSESDetailedFieldGroup.civil_engineering,
            NCSESDetailedFieldGroup.electrical_eng,
            NCSESDetailedFieldGroup.industrial_eng,
            NCSESDetailedFieldGroup.materials_engineering,
            NCSESDetailedFieldGroup.mech_e,
        ],
        NCSESFieldGroup.sci_life_sci: [
            NCSESDetailedFieldGroup.ag_sci,
            NCSESDetailedFieldGroup.bio,
        ],
        NCSESFieldGroup.math_cs: [
            NCSESDetailedFieldGroup.compsci,
            NCSESDetailedFieldGroup.math_stats,
        ],
        NCSESFieldGroup.phys_sci: [
            NCSESDetailedFieldGroup.phys_sci_other,
            NCSESDetailedFieldGroup.astronomy,
            NCSESDetailedFieldGroup.chem,
            NCSESDetailedFieldGroup.physics,
        ],
        NCSESFieldGroup.psych: [NCSESDetailedFieldGroup.psych],
        NCSESFieldGroup.social_sciences: [
            NCSESDetailedFieldGroup.area_ethnic,
            NCSESDetailedFieldGroup.social_sci_other,
            NCSESDetailedFieldGroup.polisci,
            NCSESDetailedFieldGroup.anthro,
            NCSESDetailedFieldGroup.econ,
            NCSESDetailedFieldGroup.linguistics,
            NCSESDetailedFieldGroup.sociology,
            NCSESDetailedFieldGroup.hist_sci,
        ],
        NCSESFieldGroup.geosci: [
            NCSESDetailedFieldGroup.atmospheric_sciences,
            NCSESDetailedFieldGroup.earth_sci,
            NCSESDetailedFieldGroup.oceanography,
        ],
    },
}

# Mapping between NCSES Broad Fields and NSF Broad Fields
NSF_REPORT_BROAD_FIELD_MAP = {
    NCSESFieldGroup.social_sciences: NSFBroadField.soc_behav_sci,
    NCSESFieldGroup.math_cs: NSFBroadField.math_cs,
    NCSESFieldGroup.sci_life_sci: NSFBroadField.ag_and_bio_sci,
    NCSESFieldGroup.engineering: NSFBroadField.eng,
    NCSESFieldGroup.phys_sci: NSFBroadField.phys_earth_sci,
    NCSESFieldGroup.psych: NSFBroadField.soc_behav_sci,
    NCSESFieldGroup.geosci: NSFBroadField.phys_earth_sci,
    NCSESFieldGroup.unknown: NSFBroadField.unknown,
    NCSESSciGroup.non_sci: NSFBroadField.non_stem,
    NCSESSciGroup.unknown: NSFBroadField.unknown,
}
